"use strict";
var invariant = require('invariant');
var Attrs_1 = require('./Attrs');
(function (CtrlType) {
    CtrlType[CtrlType["TITLE"] = 0] = "TITLE";
    CtrlType[CtrlType["ETC"] = 1] = "ETC";
})(exports.CtrlType || (exports.CtrlType = {}));
var CtrlType = exports.CtrlType;
;
function link(prev, next) {
    if (prev) {
        prev.next = next;
    }
    if (next) {
        next.prev = prev;
    }
}
function validateTitle(tagName, value) {
    invariant(typeof value === 'string', 'render value for <title> must be a string');
}
exports.validateTitle = validateTitle;
function validateEtc(tagName, value) {
    invariant(typeof value === 'object', "render value for <" + tagName + "> must be an object");
}
exports.validateEtc = validateEtc;
exports.validator = (_a = {},
    _a[CtrlType.TITLE] = validateTitle,
    _a[CtrlType.ETC] = validateEtc,
    _a
);
var Controller = (function () {
    /**
     * Controller constructor.
     *
     * Controller is an object that manages an element in `<head>` element.
     * If multiple controllers are bound to one element, only the last bound
     * controller is used.
     * The last bound controller is called top controller.
     *
     * When top controller is removed, previous controller of that one becomes
     * the new top controller.
     *
     * @constructs Controller
     * @param {CtrlType} type Type of controller.
     * @param {string} tagName Element tag name.
     * @param {object} attrs Object containing key-value attribute pairs, which is
     * used to find element to bind controller to.
     * @param {function} render A function that returns new element value.
     * The type of return value depends on controller type.
     * @param {object} options Controller options.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @return {Controller} a new Controller instance.
     */
    function Controller(type, tagName, attrs, render, options) {
        if (options === void 0) { options = {}; }
        invariant(typeof render === 'function', 'render function is required');
        invariant(!options.addListener || options.removeListener, 'addListener requires removeListener');
        this.type = type;
        this.attrs = new Attrs_1.default(attrs);
        this.tagName = tagName;
        this.selector = this.tagName + this.attrs.selector;
        this.render = render;
        this.addListener = options.addListener;
        this.removeListener = options.removeListener;
        this.needsToClose = !!options.close;
    }
    Controller.prototype.insertAfter = function (node) {
        link(node, this.next);
        link(this, node);
    };
    Controller.prototype.unlink = function () {
        link(this.prev, this.next);
        this.prev = undefined;
        this.next = undefined;
    };
    return Controller;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = Controller;
var _a;
//# sourceMappingURL=Controller.js.map