"use strict";
var invariant = require('invariant');
var Controller_1 = require('./Controller');
var utils_1 = require('./utils');
var OverrideListener = (function () {
    function OverrideListener(override) {
        this._controllerListeners = [];
        this._override = override;
        this._listening = false;
    }
    OverrideListener.prototype.onUpdate = function () {
        this._controllerListeners.forEach(function (listener) { return listener.onUpdate(); });
    };
    OverrideListener.prototype.addControllerListener = function (controllerListener) {
        var _this = this;
        this._controllerListeners.push(controllerListener);
        if (this._override.addListener && !this._listening) {
            this._listening = true;
            this._listener = function () { return _this.onUpdate(); };
            this._addListenerRet = this._override.addListener(this._listener);
        }
    };
    OverrideListener.prototype.removeControllerListener = function (controllerListener) {
        utils_1.removeItem(this._controllerListeners, controllerListener);
        if (this._listening && this._controllerListeners.length === 0) {
            this._override.removeListener(this._listener, this._addListenerRet);
            this._listening = false;
        }
    };
    OverrideListener.prototype.isReadyToRemove = function () {
        return !this._listening && this._controllerListeners.length === 0;
    };
    return OverrideListener;
}());
exports.OverrideListener = OverrideListener;
function emptyFunction() {
    return undefined;
}
var ControllerListener = (function () {
    function ControllerListener(_a) {
        var controller = _a.controller, overrideListener = _a.overrideListener, handler = _a.handler;
        this._overrideListener = overrideListener;
        this._controller = controller;
        this._onUpdate = handler.onUpdate;
        this._onStop = handler.onStop || emptyFunction;
        this._listening = false;
    }
    ControllerListener.prototype.onUpdate = function () {
        this._onUpdate(this._controller);
        return this;
    };
    ControllerListener.prototype.onStop = function () {
        this._onStop(this._controller);
        return this;
    };
    ControllerListener.prototype.startListening = function () {
        var _this = this;
        this._overrideListener.addControllerListener(this);
        if (this._controller.addListener) {
            this._listener = function () { return _this.onUpdate(); };
            this._addListenerRet = this._controller.addListener(this._listener);
        }
        this._listening = true;
        return this;
    };
    ControllerListener.prototype.stopListening = function () {
        this._overrideListener.removeControllerListener(this);
        if (this._controller.removeListener) {
            this._controller.removeListener(this._listener, this._addListenerRet);
        }
        this._listening = false;
        return this;
    };
    ControllerListener.prototype.isListening = function () {
        return this._listening;
    };
    ControllerListener.prototype.isReadyToRemove = function () {
        return !this.isListening();
    };
    return ControllerListener;
}());
exports.ControllerListener = ControllerListener;
var EventManager = (function () {
    function EventManager(titleHandler, etcHandler) {
        this._overrideListeners = {};
        this._controllerListeners = {};
        this._eventHandlers = (_a = {},
            _a[Controller_1.CtrlType.TITLE] = titleHandler,
            _a[Controller_1.CtrlType.ETC] = etcHandler,
            _a
        );
        var _a;
    }
    EventManager.prototype.isControllerListening = function (controller) {
        var controllerListener = this._getControllerListener(controller);
        if (!controllerListener) {
            return false;
        }
        return controllerListener.isListening();
    };
    EventManager.prototype.startListeningController = function (controller) {
        var controllerListener = this._ensureControllerListener(controller);
        controllerListener
            .onUpdate()
            .startListening();
    };
    EventManager.prototype.stopListeningController = function (controller) {
        invariant(this.isControllerListening(controller), 'Only listening controller can be stopped');
        var controllerListener = this._getControllerListener(controller);
        controllerListener
            .stopListening()
            .onStop();
    };
    EventManager.prototype.removeController = function (controller) {
        var controllerListener = this._getControllerListener(controller);
        if (!controllerListener) {
            return;
        }
        invariant(controllerListener.isReadyToRemove(), 'ControllerListener is not ready to remove');
        delete this._controllerListeners[controller.id];
    };
    EventManager.prototype.updateOverride = function (override) {
        var overrideListener = this._overrideListeners[override.id];
        if (!overrideListener) {
            return;
        }
        overrideListener.onUpdate();
    };
    EventManager.prototype.removeOverride = function (override) {
        var overrideListener = this._overrideListeners[override.id];
        if (!overrideListener) {
            return;
        }
        invariant(overrideListener.isReadyToRemove(), 'OverrideListener is not ready to remove');
        delete this._overrideListeners[override.id];
    };
    EventManager.prototype.destroy = function () {
        var _this = this;
        Object.keys(this._controllerListeners)
            .map(function (id) { return _this._controllerListeners[id]; })
            .filter(function (listener) { return listener.isListening(); })
            .forEach(function (listener) {
            listener.stopListening().onStop();
        });
    };
    EventManager.prototype._ensureOverrideListener = function (override) {
        var overrideListener = this._overrideListeners[override.id];
        if (!overrideListener) {
            overrideListener = this._overrideListeners[override.id] = new OverrideListener(override);
        }
        return overrideListener;
    };
    EventManager.prototype._getControllerListener = function (controller) {
        return this._controllerListeners[controller.id];
    };
    EventManager.prototype._ensureControllerListener = function (controller) {
        var controllerListener = this._getControllerListener(controller);
        if (controllerListener) {
            return controllerListener;
        }
        var overrideListener = this._ensureOverrideListener(controller.override);
        controllerListener = this._controllerListeners[controller.id] = new ControllerListener({
            controller: controller,
            overrideListener: overrideListener,
            handler: this._eventHandlers[controller.type],
        });
        return controllerListener;
    };
    return EventManager;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = EventManager;
//# sourceMappingURL=EventManager.js.map