import { AttrsObject } from './Attrs';
import Controller, { ControllerOptions, CtrlType, RenderFunc, AddListener, RemoveListener } from './Controller';
import Hairdresser from './Hairdresser';
export interface OverrideOptions {
    addListener?: AddListener;
    removeListener?: RemoveListener;
}
export declare type StringOrRenderFunc<T> = string | AttrsObject | RenderFunc<T>;
/**
 * Override is the unit of override and consists of a set of controllers.
 *
 * When controller is added to override, the controller overrides top controller.
 * That is, Added controller becomes new top controller and starts listenings,
 * and previous top controller stops listening.
 *
 * @class Override
 * @param {object} options Override options.
 * @return {Override} a new Override instance.
 */
declare class Override {
    id: number;
    controllers: Controller<any>[];
    addListener: AddListener;
    removeListener: RemoveListener;
    private _hairdresser;
    constructor(hairdresser: Hairdresser, options: OverrideOptions);
    /**
     * Add controller for `<title>` element.
     *
     * @memberof Override
     * @param {*} render A string which replaces `<title>` value or a function
     * that returns the title replacing string value.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @return {Override} Itself.
     */
    title(render: StringOrRenderFunc<string>, options?: ControllerOptions): this;
    /**
     * Add controller for element with given tag name.
     *
     * @memberof Override
     * @param {string} tagName Element tag name.
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    tag(tagName: string, attrs: AttrsObject, render: StringOrRenderFunc<AttrsObject>, options?: ControllerOptions): this;
    /**
     * Add controller for `<meta>` element.
     *
     * @memberof Override
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    meta(attrs: AttrsObject, render: StringOrRenderFunc<AttrsObject>, options?: ControllerOptions): this;
    /**
     * Add controller for `<link>` element.
     *
     * @memberof Override
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    link(attrs: AttrsObject, render: StringOrRenderFunc<AttrsObject>, options?: ControllerOptions): this;
    /**
     * Rerender elements of override's listening controllers.
     *
     * @memberof Override
     * @return {undefined}
     */
    update(): void;
    /**
     * Remove all override's controllers.
     *
     * When controller is top, the previous controller will become top.
     * If previous controller does not exists, the element bound to this
     * controller will be removed.
     *
     * @memberof Override
     * @return {undefined}
     */
    restore(): void;
    getController(tagName: string, attrs?: AttrsObject): Controller<any>;
    /**
     * Add controller for arbitrary element.
     *
     * @memberof Override
     * @private
     * @param {CtrlType} type Type of controller.
     * @param {string} tagName Element tag name.
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render A value that replaces element contents or a function
     * that returns the element replacing value.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @return {Override} Itself.
     */
    addController<T>(type: CtrlType, tagName: string, attrs: AttrsObject, render: StringOrRenderFunc<T>, options?: ControllerOptions): this;
}
export default Override;
