"use strict";
var invariant = require('invariant');
var Attrs_1 = require('./Attrs');
var Controller_1 = require('./Controller');
/**
 * Override is the unit of override and consists of a set of controllers.
 *
 * When controller is added to override, the controller overrides top controller.
 * That is, Added controller becomes new top controller and starts listenings,
 * and previous top controller stops listening.
 *
 * @class Override
 * @param {object} options Override options.
 * @return {Override} a new Override instance.
 */
var Override = (function () {
    function Override(hairdresser, options) {
        invariant(!options.addListener || options.removeListener, 'addListener requires removeListener');
        this._hairdresser = hairdresser;
        this.addListener = options.addListener;
        this.removeListener = options.removeListener;
        this.controllers = [];
        this.id = hairdresser._generateOverrideId();
    }
    /**
     * Add controller for `<title>` element.
     *
     * @memberof Override
     * @param {*} render A string which replaces `<title>` value or a function
     * that returns the title replacing string value.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @return {Override} Itself.
     */
    Override.prototype.title = function (render, options) {
        return this.addController(Controller_1.CtrlType.TITLE, 'title', {}, render, options);
    };
    ;
    /**
     * Add controller for element with given tag name.
     *
     * @memberof Override
     * @param {string} tagName Element tag name.
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    Override.prototype.tag = function (tagName, attrs, render, options) {
        return this.addController(Controller_1.CtrlType.ETC, tagName, attrs, render, options);
    };
    ;
    /**
     * Add controller for `<meta>` element.
     *
     * @memberof Override
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    Override.prototype.meta = function (attrs, render, options) {
        return this.tag('meta', attrs, render, options);
    };
    ;
    /**
     * Add controller for `<link>` element.
     *
     * @memberof Override
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render An object that represents new key-value attribute pairs
     * which will be added to controller's element or a function that returns
     * the object.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @param {boolean} options.close Whether to append closing tag.
     * @return {Override} Itself.
     */
    Override.prototype.link = function (attrs, render, options) {
        return this.tag('link', attrs, render, options);
    };
    ;
    /**
     * Rerender elements of override's listening controllers.
     *
     * @memberof Override
     * @return {undefined}
     */
    Override.prototype.update = function () {
        var _this = this;
        // Fire event to listening observers
        this._hairdresser._observerList.forEach(function (observer) {
            observer.onUpdate(_this);
        });
    };
    ;
    /**
     * Remove all override's controllers.
     *
     * When controller is top, the previous controller will become top.
     * If previous controller does not exists, the element bound to this
     * controller will be removed.
     *
     * @memberof Override
     * @return {undefined}
     */
    Override.prototype.restore = function () {
        var _this = this;
        // Fire event to listening observers
        this._hairdresser._observerList.forEach(function (observer) {
            observer.onPreRemoveOverride(_this);
        });
        this.controllers.forEach(function (controller) {
            if (controller.prev) {
                _this._hairdresser._topControllers[controller.selector] = controller.prev;
            }
            else {
                delete _this._hairdresser._topControllers[controller.selector];
            }
            controller.unlink();
        });
    };
    ;
    Override.prototype.getController = function (tagName, attrs) {
        var selector = tagName + (attrs ? Attrs_1.toSelector(attrs) : '');
        // Loop from tail to head to select latest inserted value.
        for (var i = this.controllers.length - 1; i >= 0; --i) {
            var controller = this.controllers[i];
            if (controller.selector === selector) {
                return controller;
            }
        }
        return null;
    };
    ;
    // TODO: Inherit document from Controller class
    /**
     * Add controller for arbitrary element.
     *
     * @memberof Override
     * @private
     * @param {CtrlType} type Type of controller.
     * @param {string} tagName Element tag name.
     * @param {object} attrs Object containing key-value attribute pairs which
     * is used to find element to bind controller to.
     * @param {*} render A value that replaces element contents or a function
     * that returns the element replacing value.
     * @param {object} options Options for element controller.
     * @param {addListener} options.addListener A function that adds controller
     * listener to event emitter.
     * @param {removeListener} options.removeListener A function that removes
     * controller listener.
     * @return {Override} Itself.
     */
    Override.prototype.addController = function (type, tagName, attrs, render, options) {
        if (options === void 0) { options = {}; }
        var renderFunc;
        if (typeof render !== 'function') {
            switch (type) {
                case Controller_1.CtrlType.TITLE: {
                    Controller_1.validateTitle(tagName, render);
                    break;
                }
                case Controller_1.CtrlType.ETC: {
                    Controller_1.validateEtc(tagName, render);
                    break;
                }
                default: {
                    break;
                }
            }
            renderFunc = function () { return render; };
        }
        else {
            renderFunc = render;
        }
        var controller = new Controller_1.default(type, tagName, attrs, renderFunc, options);
        this._hairdresser._addControllerToOverride(this, controller);
        return this;
    };
    ;
    return Override;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = Override;
//# sourceMappingURL=Override.js.map